/*
Copyright (C) 2003-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef RGCSG_H
#define RGCSG_H

#include <rgStandard.h>
#include <vector>
#include <GL/gl.h>

// XXX: Brush identifiers 0 through 1000 (inclusive) are reserved by the
//	"rgCSG" SDK. Brush identifiers 1001 through 2000 (inclusive) are
//	reserved by the "rg3DRom" SDK.
#define RGCSG_BRUSH	0
#define RGCSG_CUBE	1


/*
class rgCSGFace
{
public:
	int vert_count() const { return int( m_vert.size());}
	rgVector operator []( int i) const {
		if( (i < 0) || (i >= vert_count())) return rgVector();
		return m_vert[i];
	}

protected:
	std::vector <rgVector> m_vert;
};
*/

#define RGCSG_UNION_BRUSH		0
#define RGCSG_SUBTRACTION_BRUSH		1
#define RGCSG_INTERSECTION_BRUSH	2
#define RGCSG_DIFFERENCE_BRUSH		3

// XXX: Generic Brush
class rgCSGBrush
{
public:
	rgCSGBrush(){
		m_brush_flags = 0;
		m_operation = RGCSG_UNION_BRUSH;
		m_tex_id = -1;
		m_group_id = -1;
	}

	virtual void refresh(){}
	virtual void draw(){}
	//virtual void draw_cursor( const rgVector vc, const rgVector ec){}
	void draw_cursor( const rgVector vc, const rgVector ec) {
		GLint pmode[2];
		glGetIntegerv( GL_POLYGON_MODE, pmode);
		real_draw_cursor( vc, ec);
		glPolygonMode( GL_FRONT, pmode[0]);
		glPolygonMode( GL_BACK, pmode[1]);
	}

	bool ray_hit( const rgVector rp, const rgVector rd) const {
		rgVector ip;
		return ray_hit( rp, rd, ip);
	}
	virtual bool ray_hit( const rgVector rp, const rgVector rd,
		rgVector &ip) const {
		return false;}

	virtual rgVector position() const { return m_pos.p;}

	void set_position( const rgVector p) {
		m_pos.p = p;
		refresh();
	}

	void translate( const rgVector v) {
		m_pos.translate( v);
		refresh();
	}
	void translate( const rgMatrix &t) {
		m_pos.translate( t);
		refresh();
	}

	void rotate( const float ang, const rgVector axis) {
		m_pos.rotate( ang, axis);
		refresh();
	}
	void rotate( const rgMatrix &m) {
		m_pos.rotate( m);
		refresh();
	}

	void scale( float x, float y, float z) {
		scale( rgVector( x, y, z));}
	void scale( const rgVector v) {
		m_pos.scale( v);
		refresh();
	}

	rgCoordSystem raw_coord_system() const { return m_pos;}
	virtual rgCoordSystem coord_system() const { return m_pos;}
	void set_coord_system( rgCoordSystem c) { m_pos = c;}

	friend std::ostream &operator <<( std::ostream &o,
		const rgCSGBrush *b);
	virtual std::ostream &print( std::ostream &o) const { return o;}

	// XXX: Every derived class MUST override the following functions
	//	without exceptions. Sub-classes of sub-classes must also
	//	override these functions, otherwise confusion and inconsistency
	//	between classes can result (this is because more likely than
	//	not, no two classes will use the same version or the same save
	//	and load functionality).
	virtual int type() const { return RGCSG_BRUSH;}
	virtual int version() const { return 0;}
	virtual int save_data( std::ostream &des) const;
	virtual int load_data( std::istream &src);

	int set_group_id( int gid) { return (m_group_id = gid);}
	int set_tex_id( int tid) { return (m_tex_id = tid);}
	int group_id() const { return m_group_id;}
	int tex_id() const { return m_tex_id;}

	virtual bool selectable() const { return true;}

	virtual void gen_verts( std::vector <rgVector> &v) const { v.clear();}
	virtual void map_verts( std::vector <rgVector> &v,
		const rgCoordSystem &cs) const { v.clear();}

protected:
	rgCoordSystem m_pos;
	int m_brush_flags;
	int m_operation;
	int m_tex_id;
	int m_group_id;
	virtual void real_draw_cursor( const rgVector vc, const rgVector ec){}

private:
	rgCSGBrush( const rgCSGBrush &);
	rgCSGBrush &operator =( const rgCSGBrush &);

	int load_version_0_data( std::istream &src);
};


// XXX: Cube
class rgCSGCube : public rgCSGBrush
{
public:
	rgCSGCube();
	virtual ~rgCSGCube() { clear();}

	void clear();
	void refresh();
	void draw();
	void real_draw_cursor( const rgVector vc, const rgVector ec);
	//bool ray_hit( const rgVector rp, const rgVector rd, rgVector &ip)
	//	const;

	int type() const { return RGCSG_CUBE;}
	int version() const { return 0;}
	int save_data( std::ostream &des) const;
	int load_data( std::istream &src);

	void gen_verts( std::vector <rgVector> &v) const;
	void map_verts( std::vector <rgVector> &v, const rgCoordSystem &cs)
		const;

protected:
	// XXX: Cube is defined completely by two corner vertices/points.
	rgVector m_v[2];

	GLuint m_display_list;

	void gen_verts( rgVector *v) const;

private:
	rgCSGCube( const rgCSGCube &);
	rgCSGCube &operator =( const rgCSGCube &);

	void real_render() const;
	int load_version_0_data( std::istream &src);
};

class rgCSGWorld
{
public:
	rgCSGWorld();
	~rgCSGWorld();

	int add_brush( rgCSGBrush *b);
	void delete_brush( const int bid);
	void delete_brush( const rgCSGBrush *b);

	rgCSGBrush *brush( const int i) {
		if( !valid_brush_id( i)) return 0;
		return m_brush[i];
	}

	void clear();

	int brush_count() const { return int( m_brush.size());}
	int select_brush( rgVector p, const rgVector n,
		const float min_dis) const;

	void draw_all_brushes() {
		for( int i = 0; i < brush_count(); i++) (m_brush[i])->draw();}
	//	for( int i = 0; i < brush_count(); i++) draw_brush( i);}

	//void draw_brush( const int i) {
	//	if( !valid_brush_id( i)) return;
	//	(m_brush[i])->draw();
	//}

	bool valid_brush_id( const int i) {
		if( (i < 0) || (i >= brush_count())) return false;
		return true;
	}

protected:
	std::vector <rgCSGBrush *> m_brush;

private:
	rgCSGWorld( const rgCSGWorld &);
	rgCSGWorld &operator =( const rgCSGWorld &);
};
#endif
