/*
Copyright (C) 2004-2005 Amour Hassan
                                                                                
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
                                                                                
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
                                                                                
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
                                                                                
*/

#include "rg3DRom.h"
#include "rg3DRom_csg.h"
#include <rgPort.h>
#include <rgCSG.h>
#include <iostream>
#include <fstream>
#include <string.h>
#include <math.h>
#include "rg3DRom_file.h"

#include <rgBSP_template.cpp>

// Enable any of the below definitions to debug the sk3DRom loader.
//#define TAG_DEBUG
//#define TEXTURE_DATA_DEBUG
//#define FACE_DEBUG
//#define VERTEX_DEBUG
//#define SKELETON_NAME_DEBUG
//#define LOAD_3DROM_DEBUG
//#define SKELETON_DEBUG

/*
sk3DRomJoint::sk3DRomJoint()
{
	rcount = -1;
}
*/

sk3DRomVertex::sk3DRomVertex()
{
	weight = 1.0f;
	m_bid = -1;
	m_gid = -1;
	flags = 0;
}

sk3DRomVertex::sk3DRomVertex( const float p_x, const float p_y,
	const float p_z)
	: rgVector( p_x, p_y, p_z)
{
	weight = 1.0f;
	m_bid = -1;
	m_gid = -1;
	flags = 0;
}

sk3DRomVertex::sk3DRomVertex( const rgVector &v)
{
	x = v.x;
	y = v.y;
	z = v.z;
	weight = 1.0f;
	m_bid = -1;
	m_gid = -1;
	flags = 0;
}

sk3DRomVertex &sk3DRomVertex::operator =( const rgVector &v)
{
	x = v.x;
	y = v.y;
	z = v.z;
	return *this;
}

sk3DRomFace::sk3DRomFace()
{
	flags = 0;
	gid = -1;
	memset( li, 255, sizeof( li));	// Pure white and fully opaque
}

/*
sk3DRomTexture::sk3DRomTexture( const sk3DRomTexture &n)
{
	if( n.tex) {
		tex = new rgGLTexture( *(n.tex));
		fname = n.fname;
		flags = n.flags;
	} else {
		tex = 0;
	}
}

sk3DRomTexture::~sk3DRomTexture()
{
	delete tex;
}

sk3DRomTexture &sk3DRomTexture::operator =( const sk3DRomTexture &n)
{
	if( n.tex) {
                tex = new rgGLTexture( *(n.tex));
		fname = n.fname;
		flags = n.flags;
        } else {
                tex = 0;
        }

	return *this;
}
*/

sk3DRomBoneFrame::sk3DRomBoneFrame()
{
	msb = -1;
	mmb = -1;
	cms = -1;
}

sk3DRomSkelFrame::sk3DRomSkelFrame()
{
	edata = 0;
	memset( fname, 0, SK3DROM_SKEL_FRAME_NAME_LEN);
}

// Description: Sets the pivot joint of a given bone "bid" to be "jid".
//
// XXX: Mounted bones are also affected by a pivot joint change. Mounted
//	bones can be updated by saving a list of all mounted slave bones.
//	Then, all slaves are unmounted, the pivot joint is changed, and then
//	the list of slave bones are remounted with the new pivot joint. All
//	skeleton frames must be if they have mounted bones that will be
//	affected.
//
// TODO: Add support for changing pivot joints with mounted bones.
void sk3DRomModel::set_pivot_joint( const int bid, const short jid)
{
	if( (bid < 0) || (bid >= int( skel.b.size()))) return;
	// The coordinate system of one or more bones will be changed.
	// Therefore all data dependant on the affected coordinate systems
	// must be "unbound" and "rebound" to the new coordinate systems.

	// "Un-bind" all vertices from all bones that will be affected.
	unbind_sub_skel_verticies( bid);	

	// Set the pivot joint for the selected bone.
	skel.set_pivot_joint( bid, jid);

	// "Re-bind" all vertices from all bones that were affected.
	bind_sub_skel_verticies( bid);
}

void sk3DRomModel::bind_sub_skel_verticies( const short bid)
{
	// "Bind" all of the vertices controlled by this bone.
	bind_bone_verticies( bid);

	// "Bind" all vertices controlled by slave bones.
	sk3DRomBone &b = skel.b[bid];
	const short istop = short( b.link.size());
	for( short i = 0; i < istop; i++)
		bind_sub_skel_verticies( b.link[i].bid);
}

void sk3DRomModel::unbind_sub_skel_verticies( const short bid)
{
	// "Unmap" all of the vertices controlled by this bone.
	unmap_bone_vertices( bid, 0);

	// "Unbind" all vertices controlled by slave bones.
	sk3DRomBone &b = skel.b[bid];
	const short istop = short( b.link.size());
	for( short i = 0; i < istop; i++)
		unbind_sub_skel_verticies( b.link[i].bid);
}

void hlp_init_rcs_for_all_frames( const short bid, sk3DRomSkel &s)
{
	s.skout.b[bid].rcs = s.b[bid].jcs;
	const int istop = int( s.sf.size());
	for( int i = 0; i < istop; i++) s.sf[i].b[bid].rcs = s.b[bid].jcs;
}

void sk3DRomSkel::set_pivot_joint( const int bid, const short jid)
{
	// Check parameters.
	if( bid < 0) return;
	if( bid >= int( b.size())) return;

	// If pivot joint is already set, then do nothing.
	if( jid == b[bid].pj) return;

	// Set the joint id.
	switch( jid) {
	case 1: b[bid].pj = 1; break;
	default: b[bid].pj = 0; break;
	}

	// Set the coordinate system for the pivot joint.
	rgCoordSystem mcs( get_full_bone_jcs( b[bid].mid));
	rgCoordSystem tcs, ecs;
	ecs = tcs = mcs.map_csys( b[bid].jcs);

	// Change the coordinate system.
	tcs.p = b[bid].get_pivot_joint();
	// Since pivot joint changed, the bone vector (up) must be the
	// reverse. Also, since the "side" vector is a result of "up" cross
	// "eye", it must also be the reverse.
	tcs.up = -1.0f * tcs.up;
	tcs.side = -1.0f * tcs.side;

	// Change the coordinate system of all bone information.
	b[bid].ja = tcs.change_point_csys( b[bid].a);
	b[bid].jb = tcs.change_point_csys( b[bid].b);
	b[bid].jcs = mcs.change_csys_csys( tcs);

	// Update the rcs of this bone for all frames.
	hlp_init_rcs_for_all_frames( bid, *this);

	// All (and only) "immediate" slave bones need to be notified of the
	// pivot joint change.
	mcs = mcs.map_csys( b[bid].jcs);
	const short istop = short( b[bid].link.size());
	for( short i = 0; i < istop; i++) {
		const short k = b[bid].link[i].bid;
		b[k].jcs = mcs.change_csys_csys( ecs.map_csys( b[k].jcs));
		// Update the rcs of this bone for all frames.
		hlp_init_rcs_for_all_frames( k, *this);
	}
}

void sk3DRomSkel::reset_all_skel_frames()
{
	// Update all skeleton frames
	const short istop = short( sf.size());
	const short jstop = short( b.size());
	short j;
	for( short i = 0; i < istop; i++) {

		for( j = 0; j < jstop; j++)
			sf[i].b[j].rcs = b[j].jcs;
	}
}

void sk3DRomSkel::reset_skel_frame( const int sid)
{
	sk3DRomSkelFrame *s = &skout;
	if( (sid >= 0) && (sid < frame_count())) s = &sf[ sid];
	if( s->b.size() != b.size()) init_skel_frame( *s);
	for( int i = 0; i < int( b.size()); i++) s->b[i].rcs = b[i].jcs;
}

void sk3DRomSkel::get_bone_coords( const short bid, const short sid,
	rgVector &p_a, rgVector &p_b)
{
	if( (bid < 0) || (bid >= int( b.size()))) return;
	if( (sid < 0) || (sid >= int( sf.size())))
		get_bone_coords( bid, skout, p_a, p_b);
	else get_bone_coords( bid, sf[sid], p_a, p_b);
}

void sk3DRomSkel::get_bone_coords( const short bid,
	const sk3DRomSkelFrame &p_sf, rgVector &p_a, rgVector &p_b)
{
	rgCoordSystem cs = get_full_bone_rcs( bid, p_sf);
	p_a = cs.map_point( b[bid].ja);
	p_b = cs.map_point( b[bid].jb);
}

void sk3DRomSkel::build_master_list()
{
	const short istop = short( b.size());
	for( short i = 0; i < istop; i++) build_bone_master_list( i);
}

void sk3DRomSkel::build_bone_master_list( const int bid)
{
	b[bid].mlist.clear();
	short j = b[bid].mid;
	while( j >= 0) {
		b[bid].mlist.push_back( j);
		j = b[j].mid;
	}
}

rgCoordSystem sk3DRomSkel::get_full_bone_jcs( const short bid) const
{
	// Apply the master jcs's (however many there are).
	rgCoordSystem cs;

	// If an invalid bone id is given, then return the standard
	// coordinate system.
	if( (bid < 0) || (bid >= int( b.size()))) return cs;

	const short istart = short( b[bid].mlist.size() - 1);
	for( short i = istart; i >= 0; i--)
		cs = cs.map_csys( b[ b[bid].mlist[i]].jcs);

	// Apply the bone's jcs.
	return cs.map_csys( b[bid].jcs);
}

//rgCoordSystem sk3DRomSkel::get_full_bone_rcs( const short bid, const short sfid)
//{
//	return get_full_bone_rcs( bid, sf[sfid]);
//}

//void sk3DRomSkel::get_full_bone_rcs( const short bid,
//	const short sfid, rgCoordSystem &rcs) const
//{
//	rcs = get_full_bone_rcs( bid, sfid);
//}

void sk3DRomSkel::rotate_rcs(
                const short bid,	// The bone id
                const short sid,	// The skeleteon frame id
		const float ang,
		const rgVector axis)	// Axis of rotation (normalized)
{
	if( sid >= short( sf.size())) return;
	if( sid >= 0) rotate_rcs( bid, sf[sid], ang, axis);
	else rotate_rcs( bid, skout, ang, axis);
}

void sk3DRomSkel::rotate_rcs( const short bid, sk3DRomSkelFrame &p_sf,
	const float ang, const rgVector axis)
{
	if( bid < 0) return;
	if( bid >= short( b.size())) return;

	rgMatrix m;
	m.rotate( ang, axis);

	rgCoordSystem mcs;
	rgCoordSystem rcs( get_full_bone_rcs( bid, p_sf, mcs));

	rcs.side = m * rcs.side;
	rcs.up = m * rcs.up;
	rcs.eye = m * rcs.eye;

	p_sf.b[bid].rcs.side = mcs.change_vector_csys( rcs.side);
	p_sf.b[bid].rcs.up = mcs.change_vector_csys( rcs.up);
	p_sf.b[bid].rcs.eye = mcs.change_vector_csys( rcs.eye);
}

void sk3DRomSkel::reset_bone( const short bid, const short sid)
{
	if( (bid < 0) || (bid >= int( b.size()))) return;
	if( (sid < 0) || (sid >= int( sf.size()))) reset_bone( bid, skout);
	else reset_bone( bid, sf[ sid]);
}

void sk3DRomSkel::reset_bone_and_slaves( const short bid, const short sid)
{
	if( (sid < 0) || (sid >= int( sf.size())))
		reset_bone_and_slaves( bid, skout);
	else reset_bone_and_slaves( bid, sf[ sid]);
}

void sk3DRomSkel::reset_bone_and_slaves( const short bid,
	sk3DRomSkelFrame &p_sf)
{
	if( (bid < 0) || (bid >= int( b.size()))) return;
	reset_bone( bid, p_sf);
	const int istop = int( b[bid].link.size());
	for( int i = 0; i < istop; i++)
		reset_bone_and_slaves( b[bid].link[i].bid, p_sf);
}




sk3DRomModel::sk3DRomModel()
{
	m_face_list = 0;

	tv = 0;
	tv_array_element_count = 0;

	read_custom_data = 0;
	pre_save = 0;
	write_custom_data = 0;

	custom_tag_memory_length = 0;
	custom_tag_memory_storage = 0;

	render_format = SK3DROM_RENDER_FORMAT_V4N3E1T2E6;

	m_brush_cursor = 0;//new sk3DRomCSGCube();
	m_csg = new rgCSGWorld();

	m_gl_vertex_array = 0;
}

sk3DRomModel::~sk3DRomModel()
{
	clear();
	if( m_csg) delete m_csg;
}

void sk3DRomModel::clear()
{
	delete_vertex_array();

	if( m_face_list) { delete m_face_list; m_face_list = 0;}

	// Delete the vertex transformation array (if present).
	if( tv) {
		delete [] tv;
		tv = 0;
		tv_array_element_count = 0;
	}

	v.clear();
	f.clear();
	group.clear();
	texture.clear();
	bone_less_verts.clear();

	// Clear all main skeleton data
	skel.b.clear();
	skel.rb.clear();
	skel.sf.clear();
	skel.seq.clear();

	// Delete all BSP data.
	clear_bsp_data();

	// Clear the custom tag memory.
	clear_custom_tag_memory();

	m_brush_cursor = 0;
	m_csg->clear();
}

void sk3DRomModel::gen_master_joint_links()
{
	sk3DRomBone *b;
	const short istop = short( skel.b.size());
	short i, j, jstop;
	for( i = 0; i < istop; i++) {
		
		b = &(skel.b[i]);
		
		// Inform all slave bones of their master joint
		jstop = short( b->link.size());
		for( j = 0; j < jstop; j++) {
			///(skel.b[ (b->link[j]).bid]).mjid = (b->link[j]).jid;

			// Set the master joint vertex
			//switch( (b->link[j]).jid) {
			//case 0: (skel.b[ (b->link[j]).bid]).c = b->a; break;
			//case 1: (skel.b[ (b->link[j]).bid]).c = b->b; break;
			//}
		}
	}
}

void sk3DRomModel::gen_flat_shade_normals()
{
	if( !tv) return;
	const int istop = int( f.size());
	int i;
	rgVector tri[3];
	rgFloat *tv_ptr = tv;
	switch( render_format) {
	case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
		for( i = 0; i < istop; i++) {
			tv_ptr = tv + (int( (f[i]).vid[0]) << 4);
			tri[0].x = tv_ptr[0];
			tri[0].y = tv_ptr[1];
			tri[0].z = tv_ptr[2];

			tv_ptr = tv + (int( (f[i]).vid[1]) << 4);
			tri[1].x = tv_ptr[0];
			tri[1].y = tv_ptr[1];
			tri[1].z = tv_ptr[2];

			tv_ptr = tv + (int( (f[i]).vid[2]) << 4);
			tri[2].x = tv_ptr[0];
			tri[2].y = tv_ptr[1];
			tri[2].z = tv_ptr[2];

			(f[i]).n[0] = (f[i]).n[1] = (f[i]).n[2] =
				(rgCalcTriangleNormal( tri[0], tri[1], tri[2])
					.normalize());
		}
		break;
	}
}

/*
bool inline hlp_faces_share_all_verts( const sk3DRomFace &a,
	const sk3DRomFace &b)
{
	if( a.vid[0] == b.vid[0]) {
		if(	(a.vid[1] == b.vid[1]) &&
			(a.vid[2] == b.vid[2]))
			return true;
		if(	(a.vid[1] == b.vid[2]) &&
			(a.vid[2] == b.vid[1]))
			return true;

	} else if( a.vid[0] == b.vid[1]) {
		if(	(a.vid[1] == b.vid[0]) &&
			(a.vid[2] == b.vid[2]))
			return true;
		if(	(a.vid[1] == b.vid[2]) &&
			(a.vid[2] == b.vid[0]))
			return true;

	} else if( a.vid[0] == b.vid[2]) {
		if(	(a.vid[1] == b.vid[1]) &&
			(a.vid[2] == b.vid[0]))
			return true;
		if(	(a.vid[1] == b.vid[0]) &&
			(a.vid[2] == b.vid[1]))
			return true;
	}
	return false;
}
*/

rgVector hlp_get_smooth_vert_normal( sk3DRomModel &mdl, const int fid,
	const int vid, const int fcount, rgVector *fn)
{
	const rgVector norm( mdl.face_normal( fid));
	rgVector n;
	for( int i = 0; i < fcount; i++) {
		// If any of the face's vertices are "vid", then include this
		// face's normal in the smoothing process.
		if(	((vid == mdl.f[i].vid[0]) ||
			(vid == mdl.f[i].vid[1]) ||
			(vid == mdl.f[i].vid[2])) &&
			(norm.dot( mdl.face_normal( i)) > -0.9999))
			n = n + fn[i];
	}
	return (n.normalize());
}

void sk3DRomModel::gen_smooth_shade_normals()
{
	// FIXME: This is a very slow function. Though it was never intended
	//	to be fast. If there is time, try and optimize this function.
	if( !tv) return;
	if( f.empty()) return;

	// FIXME: Code reuse for ease of programming, however, program a
	//	custom smooth shade normal function for speed.
	gen_flat_shade_normals();

	// Save the face normals in a temporary array.
	rgTempArray <rgVector> fn( f.size());
	if( !(fn.data)) return;
	const int istop = int( f.size());
	int i;
	for( i = 0; i < istop; i++) fn.data[i] = f[i].n[0];

	// Generate the smooth normals now.
	for( i = 0; i < istop; i++) {

		// Only smoothen the face if it does not have it's "force
		// flat shade" bit set.
		if( !((f[i].flags) & SK3DROM_FACE_FLAT_BIT)) {
			f[i].n[0] = hlp_get_smooth_vert_normal( *this, i,
				f[i].vid[0], istop, fn.data);

			f[i].n[1] = hlp_get_smooth_vert_normal( *this, i,
				f[i].vid[1], istop, fn.data);

			f[i].n[2] = hlp_get_smooth_vert_normal( *this, i,
				f[i].vid[2], istop, fn.data);
		}
	}
}

bool sk3DRomModel::create_transformation_array()
{
	//dbg_tv_check_leak();

	// If there are no vertices, then delete the transformation array.
	if( v.empty()) {
		if( tv) { delete [] tv; tv = 0;}
		tv_array_element_count = 0;
		return false;
	}

	// Determine the optimal size of the array.
	const int n = int( v.size());
	//int k = tv_array_element_count;
	/*
	int k = 0;
	{
		int i = int( v.size());
		while( i >= 1) { k++; i >>= 1;}
	}
	k = int( pow( 2, k));
	*/
	
	int k = 2;
	// Reduce array size to save memory.
	while( n < (k >> 1)) k >>= 1;

	// Increase array size to save more elements.
	while( n >= k) k <<= 1;

	// The array has a minimum size of two elements.
	if( k < 2) k = 2;

	// If the array is already at the proper size, then do nothing.
	if( k == tv_array_element_count) return true;

	// Resize the array now.
	if( tv) { delete [] tv; tv = 0;}
	tv_array_element_count = k;
	switch( render_format) {
	case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
		tv = new rgFloat[ tv_array_element_count << 4];
		//tv = new rgFloat[ (tv_array_element_count << 4) + 256];
		//{
		//	rgFloat *p = tv + (k << 4);
		//	for( int i = 0; i < 256; i++) p[i] = i;
		//}
		//dbg_tv_check_leak();
		break;
	}

	if( !tv) { tv_array_element_count = 0; return false;}
	return true;
}

void sk3DRomModel::dbg_tv_check_leak()
{
	if( !tv) return;
	rgFloat *p = tv + (tv_array_element_count << 4);
	for( int i = 0; i < 256; i++) {
		if( p[i] != i) {
			std::cerr << "tv leak at \"i\" = " << i <<
				", float = " << p[i] << "\n";
			*((int *)0) = 0;
		}
	}
}

void sk3DRomModel::transform_skel_frame( const rgMatrix *m, int sfid,
	sk3DRomDrawContext *dc)
{
	// If an invalid skeleton frame was given, then simply transform
	// the default skeleton.
	if( (sfid < 0) || (sfid >= int( skel.sf.size()))) {
		if( dc) transform_skel_frame( m, dc->m_skout, dc);
		else transform_skel_frame( m, skel.skout, dc);
		return;
	}
	transform_skel_frame( m, skel.sf[ sfid], dc);
}

void sk3DRomModel::transform_skel_frame(
	const rgMatrix *m,
	const sk3DRomSkelFrame &s,
	sk3DRomDrawContext *dc)
{
	if( s.b.size() != skel.b.size()) return; // Skels must be compatible

	// XXX: Choose a vertex array to use.
//	float *va;
//	int va_stride;
//	if(	dc &&
//		(dc->m_use_vertex_array) &&
//		(dc->m_transform_direct) &&
//		m_gl_vertex_array) {
//		if( !(va = m_gl_vertex_array->vertex_array_start())) return;
//		if( (va_stride = m_gl_vertex_array->vertex_stride()) <= 0)
//			return;
//
//	} else {
		// If there is no transformation array, then try to create one.
		if( !tv) create_transformation_array();
		if( !tv) return;
//		va = tv;
//		va_stride = tv_vertex_stride();
//	}

	rgMatrix mat;
	if( dc) mat = dc->mat();
	else mat = skel.m_draw_context.mat();
	if( m) mat = (*m) * mat;

	const int istop = int( skel.b.size());
	for( int i = 0; i < istop; i++) {
		// Get the bones complete coordinate system first.
		// FIXME: It would be more efficient to transform bones
		//	starting with the master bone and working on down to
		//	the slave bones. This is because "get_full_bone_rcs"
		//	performs many redundant calculations as multiple bones
		//	can have the same "master list/path".
		rgCoordSystem jcs;
		rgCoordSystem cs;
		{
			if( dc) {
				cs = dc->get_full_bone_rcs( skel, i, s, jcs);
				dc->set_fcs( i, cs);

			} else {
				cs = skel.m_draw_context.get_full_bone_rcs(
					skel, i, s, jcs);
				skel.b[i].fcs = cs;
			}
			jcs = jcs.map_csys( skel.b[i].jcs);
		}
		// FIXME: Create two matrices to store both cs and jcs with
		// the draw_context matrix applied.
		const rgMatrix jmat( mat * (jcs.get_mapping_matrix()));
		const rgMatrix cmat( mat * (cs.get_mapping_matrix()));

		// Transform/map all vertices controlled by this bone.
		//const sk3DRomBone *b = &(skel.b[i]);
		const std::vector <short> &vid = (skel.b[i]).vid;
		for( int j = 0; j < int( vid.size()); j++) {
			const sk3DRomVertex &p = v[ vid[j]];
			const rgVector g( cmat.pointMul( p));
			const rgVector h( jmat.pointMul( p));
			const float w = p.weight;
			const float iw = 1.0f - w;
				
			rgFloat *tv_ptr = tv_vertex_offset( int( vid[j]));
//			rgFloat *tv_ptr = va + (int( vid[j]) * va_stride);
			tv_ptr[0] = (g.x * w) + (h.x * iw);
			tv_ptr[1] = (g.y * w) + (h.y * iw);
			tv_ptr[2] = (g.z * w) + (h.z * iw);
		}
	}

	// Finally, copy all bone-less vertices (if there are any).
	for( int j = 0; j < int( bone_less_verts.size()); j++) {
		const sk3DRomVertex &p = mat.pointMul( v[ bone_less_verts[j]]);
		rgFloat *tv_ptr = tv_vertex_offset( bone_less_verts[j]);
//		rgFloat *tv_ptr = va + (bone_less_verts[j] * va_stride);
		tv_ptr[0] = p.x;
		tv_ptr[1] = p.y;
		tv_ptr[2] = p.z;
	}
}

void sk3DRomModel::map_bone_verticies_and_slaves( const short bid,
	const short sfid, rgCoordSystem *mcs)
{
	sk3DRomBone &b = skel.b[bid];
	sk3DRomBoneFrame &f = (skel.sf[ sfid]).b[bid];

	// Apply the master's coordinate system (if any)
	rgCoordSystem cs;	// "cs" is the standard coord sys. by default.
	if( mcs) (*mcs).map_coord_system( cs, f.rcs);
	else cs = f.rcs;

	//std::cerr << cs << "\n";

	// Transform all vertices linked to the bone now
	transform_bone_verticies( bid, sfid, cs);

	// Recursivley transform all slave bones.
	const short istop = short( b.link.size());
	for( short i = 0; i < istop; i++) {
		map_bone_verticies_and_slaves( (b.link[i]).bid, sfid,
			&cs);
	}
}

void sk3DRomModel::map_verticies_to_skeleton()
{
	const short jstop = short( skel.b.size());
	for( short j = 0; j < jstop; j++) bind_bone_verticies( j);
}

void sk3DRomModel::transform_bone_verticies( const short bid,
	const short sfid, rgCoordSystem &cs)
{
	sk3DRomBone &b = skel.b[bid];

	const short istop = short( b.vid.size());
	std::vector <short> &u = b.vid;
	short i;
	rgVector p;
	rgFloat *tv_ptr = tv;
	switch( render_format) {
	case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
		for( i = 0; i < istop; i++) {
			//tv[ u[i]] = cs.map_point( v[ u[i]]);
			p = cs.map_point( v[ u[i]]);
			tv_ptr = tv + (int( u[i]) << 16);
			tv_ptr[0] = p.x;
			tv_ptr[1] = p.y;
			tv_ptr[2] = p.z;
			//exit(0);
			//std::cout << u[i] << "\t: v = " << v[ u[i]] <<
			//	", tv = " << tv[ u[i]] << "\n";
		}
		break;
	}
}

void sk3DRomModel::transform_bone_verticies( const short bid,
	const short sfid, rgMatrix &p_m)
{
	sk3DRomBone &b = skel.b[bid];

	const short istop = short( b.vid.size());
	std::vector <short> &u = b.vid;
	rgVector p;
	rgFloat *tv_ptr = tv;
	short i;
	switch( render_format) {
	case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
		for( i = 0; i < istop; i++) {
			//tv[ u[i]] = p_m.pointMul( v[ u[i]]);
			p = p_m.pointMul( v[ u[i]]);
			tv_ptr = tv + (int(u[i]) << 16);
			tv_ptr[0] = p.x;
			tv_ptr[1] = p.y;
			tv_ptr[2] = p.z;
		}
		break;
	}
}

void sk3DRomModel::check_faces()
{
	const int istop = int( f.size());
	for( int i = 0; i < istop; i++) {
		if( f[i].tid >= int( texture.size())) f[i].tid = -1;
	}
}


#define hlp_read_tag sk3DRom_read_tag
int sk3DRom_read_tag( std::istream &inf, sk3DRomTag &t)
{
	t.mode = read_word( inf);
	t.size = read_dword( inf);
	if( inf.fail()) return -1;
	if( t.size <= 0) return -2;

	#ifdef TAG_DEBUG
	std::cerr << "Tag { Mode = ";
	switch( t.mode) {
	case sk3DROM_MODEA: std::cerr << "ModeA"; break;
	case sk3DROM_MODEB: std::cerr << "ModeB"; break;
	case sk3DROM_MODEC: std::cerr << "ModeC"; break;
	case sk3DROM_VERTEX: std::cerr << "Vertex"; break;
	case sk3DROM_FACE: std::cerr << "Face"; break;
	case sk3DROM_TEXTURE: std::cerr << "Texture"; break;
	case sk3DROM_TSTRIP: std::cerr << "TStrip"; break;
	case sk3DROM_TEXINFO: std::cerr << "TexInfo"; break;
	case sk3DROM_SKELETON: std::cerr << "Skeleton"; break;
	case sk3DROM_SKELFRAMES: std::cerr << "SkelFrames"; break;
	case sk3DROM_VERTMODNAMES: std::cerr << "VertModNames"; break;
	default: std::cerr << "Uknown (" << t.mode << ")"; break;
	}

	std::cerr << ", Size = " << t.size << "}\n";
	#endif

	return 0;
}

void sk3DRomSkel::gen_absolute_joint_verts()
{
	rgCoordSystem cs;
	const int istop = int( b.size());
	for( int i = 0; i < istop; i++) {
		cs = get_full_bone_jcs( i);
		b[i].a = cs.map_point( b[i].ja);
		b[i].b = cs.map_point( b[i].jb);
	}
}


rgVector sk3DRom_get_bone_vector( rgVector a, rgVector b, sk3DRomBone &bone)
{
	switch( bone.pj) {
	case 0: return b - a;	// Joint "a" is the pivot
	case 1: return a - b;	// Joint "b" is the pivot
	default: return rgVector();	// No pivot, return NULL vector
	}
}

rgVector sk3DRom_get_bone_vector( sk3DRomBone &b)
{
	switch( b.pj) {
	case 0: return b.b - b.a;	// Joint "a" is the pivot
	case 1: return b.a - b.b;	// Joint "b" is the pivot
	default: return rgVector();	// No pivot, return NULL vector
	}
}

/*
rgBasis sk3DRomModel::get_bone_basis( sk3DRomBone &b)
{
	// If no master bone basis, then return the default X, Y, Z basis.
	if( b.bone_basis <= 0) return rgBasis();

	// Else, return the master bone's support basis.
	return (skel.b[ b.bone_basis]).support_basis;
}
*/

void sk3DRomModel::get_bone_angle( rgVector a, rgVector b, sk3DRomBone &bone,
	float &fang, float &sang)
{
	//rgVector u = sk3DRom_get_bone_vector( bone);
	rgVector u = sk3DRom_get_bone_vector( a, b, bone);

	//rgBasis basis = get_bone_basis( bone);
	rgBasis basis;
	basis.side = bone.jcs.side;
	basis.up = bone.jcs.up;
	basis.eye = bone.jcs.eye;
	const float t = u.dot( basis.up);

	// Get "side" angle
	sang = t;
	{
		const float temp = u.dot( basis.side);
		if( temp < 0) sang = RG_360_DEG_IN_RAD - sang;
	}

	// Get "forward" angle
	fang = t;
	{
		const float temp = u.dot( basis.eye);
		if( temp < 0) fang = RG_360_DEG_IN_RAD - fang;
	}
}


sk3DRomVertex *sk3DRomModel::add_vertex( const rgVector p_v)
{
	v.push_back( p_v);
	create_transformation_array();

	// Add the newly created vertex to the bone-less list.
	bone_less_verts.push_back( v.size() - 1);
	return &(v[ v.size() - 1]);
}

void sk3DRomModel::duplicate_vertices( const int *vlist, int count)
{
	if( !vlist) return;
	if( count <= 0) return;
	sk3DRomVertex *p;
	const int offset = int( v.size());
	int i = count;
	while( i--) {
		p = add_vertex( v[ *vlist]);
		add_vertex_to_group( int( v.size()) - 1, v[ *vlist].gid());
		p->m_bid = v[ *vlist].bid();
		vlist++;
	}
	int *tmp = new int[ count];
	for( i = 0; i < count; i++) tmp[i] = i + offset;
	unbind_vertices( tmp, count);
	delete [] tmp;
}

bool sk3DRomModel::triangle_in_mesh( const int fid, const int *vlist,
	const int count, int *vid) const
{
	int j = 0;
	for( int i = 0; i < count; i++) {

		if( f[fid].vid[0] == vlist[i]) { vid[0] = i; j |= 0x1;}
		else if( f[fid].vid[1] == vlist[i]) { vid[1] = i; j |= 0x2;}
		else if( f[fid].vid[2] == vlist[i]) { vid[2] = i; j |= 0x4;}
		// Check if complete triangle was found.
		if( (0x7 == (j & 0x7))) return true;
	}
	return false;
}

void sk3DRomModel::duplicate_mesh( const int *vlist, int count)
{
	if( !vlist) return;
	if( count < 0) return;
	const int k = int( v.size());	// First index of duplicated verts
	duplicate_vertices( vlist, count);
	if( count < 3) return;	// Need at least 3 to get one triangle

	int vid[3];
	// XXX WARNING: "f.size()" may/will increase during the loop.
	//	Therefore, must place its current size in a constant.
	const int istop = int( f.size());
	for( int i = 0; i < istop; i++) {
		if( triangle_in_mesh( i, vlist, count, vid)) {
			add_triangle( vid[0] + k, vid[1] + k, vid[2] + k);
		}
	}
}

sk3DRomFace *sk3DRomModel::add_triangle( const int v1, const int v2,
	const int v3)
{
	// XXX: Do not add the triangle if the vertex indicies are invalid.
	if( (v1 < 0) || (v1 > int( v.size()))) return 0;
	if( (v2 < 0) || (v2 > int( v.size()))) return 0;
	if( (v3 < 0) || (v3 > int( v.size()))) return 0;

	f.push_back( sk3DRomFace());
	sk3DRomFace *temp = &(f[ f.size() - 1]);
	temp->vid[0] = v1;
	temp->vid[1] = v2;
	temp->vid[2] = v3;
	return temp;
}

void sk3DRomModel::delete_face( const int fid)
{
	if( fid < 0) return;
	const int last = int( f.size() - 1);
	if( fid > last) return;

	// NOTE: The user/modler/caller should not be using BSP data while
	//	editing the model anyway, therefore delete all bsp data without
	//	performing any checks.
	clear_bsp_data();

	// Notify all entities referencing the face that it has been deleted.
	notify_pages_of_face_delete( fid);

	// If face to delete is not the last face, then swap it with the last
	// face, and delete "fid". Also, notfiy all entities referencing the
	// face that it has been swaped.
	if( fid != last) {
		notify_pages_of_face_swap( last, fid);
		f[fid] = f[last];
	}

	f.erase( f.begin() + last);
}

inline void hlp_notify_faces_of_vertex_delete( const int vid,
	sk3DRomModel &mdl)
{
	// NOTE: Checking faces in backwards order is an optimization.
	std::vector <sk3DRomFace> &f = mdl.f;
	int i = int( f.size());
	while( i--) {
		if(	(vid == f[i].vid[0]) || (vid == f[i].vid[1]) ||
			(vid == f[i].vid[2])) mdl.delete_face( i);
	}
}

inline void hlp_notify_bone_of_vertex_delete( const int vid,
	const int bid, sk3DRomSkel &skel)
{
	if( bid < 0) return;
	sk3DRomBone &b = skel.b[bid];
	const int istop = int( b.vid.size());
	// NOTE: There is only ONE reference per owned vertex
	//	in the bones vertex list "vid".
	for( int i = 0; i < istop; i++) {
		if( vid == b.vid[i]) {
			b.vid.erase( b.vid.begin() + i);
			return;
		}
	}
}

inline void hlp_notify_faces_of_vertex_swap( const int vid_old,
	const int vid_new, std::vector <sk3DRomFace> &f)
{
	for( int i = 0; i < int( f.size()); i++) {
		if( (vid_old == f[i].vid[0])) f[i].vid[0] = vid_new;
		if( (vid_old == f[i].vid[1])) f[i].vid[1] = vid_new;
		if( (vid_old == f[i].vid[2])) f[i].vid[2] = vid_new;
	}
}

inline void hlp_notify_bone_of_vertex_swap( const int vid_old,
	const int vid_new, const int bid, sk3DRomSkel &skel)
{
	if( bid < 0) return;
	sk3DRomBone &b = skel.b[bid];
	const int istop = int( b.vid.size());
	// NOTE: There is only ONE reference per owned vertex
	//	in the bones vertex list "vid".
	for( int i = 0; i < istop; i++) {
		if( vid_old == b.vid[i]) {
			b.vid[i] = vid_new;
			return;
		}
	}
}

void sk3DRomModel::add_vertex_to_bone_less_list( const int vid)
{
	for( int i = 0; i < int( bone_less_verts.size()); i++)
		if( vid == bone_less_verts[i]) return;
	bone_less_verts.push_back( vid);
}

void sk3DRomModel::remove_vert_from_bone_less_list( const int vid)
{
	for( int i = 0; i < int( bone_less_verts.size()); i++) {
		if( vid == bone_less_verts[i]) {
			bone_less_verts.erase( bone_less_verts.begin() + i);
			return;
		}
	}
}

void sk3DRomModel::delete_vertex( const int vid)
{
	if( vid < 0) return;
	const int last = int( v.size()) - 1;
	if( vid > last) return;

	// If the vertex is bone less, then remove it from the bone less list.
	if( v[ vid].bid() < 0) remove_vert_from_bone_less_list( vid);

	// Notify all entities referencing the given vertex that it will be
	// deleted.
	hlp_notify_faces_of_vertex_delete( vid, *this);
	hlp_notify_bone_of_vertex_delete( vid, v[vid].bid(), skel);

	// If the vertex to delete is not the last vertex in the list, then
	// swap it with the last vertex, and delete vertex "vid".
	if( vid != last) {

		// Notify all faces, bones and groups that the last vertex
		// will be swaped. The last vertex will be in position
		// "vid", while "vid" will be deleted.
		hlp_notify_faces_of_vertex_swap( last, vid, f);
		hlp_notify_bone_of_vertex_swap( last, vid, v[last].bid(), skel);

		// Swap this vertex with the last vertex.
		v[vid] = v[last];
		if( v[vid].bid() < 0) add_vertex_to_bone_less_list( vid);
		remove_vert_from_bone_less_list( last);
	}

	// The last vertex is always deleted.
	v.erase( v.begin() + last);

	// As a last step update the transformation array (if neccasary).
	create_transformation_array();
}

int sk3DRomModel::ortho_select_tv_triangle( rgVector p, const rgVector n)
{
	if( !tv) return -1;

	p = rgMakePlanar( p, n);
	tri_sel_cache.start_selection();
	const int istop = int( f.size());
	int i;
	for( i = 0; i < istop; i++) {

		if( !triangle_is_hidden( i)) {

			rgVector t[3];
			get_tv_face_verticies( i, t);
			t[0] = rgMakePlanar( t[0], n);
			t[1] = rgMakePlanar( t[1], n);
			t[2] = rgMakePlanar( t[2], n);

			if( rgPlanarPointInTriangle( t[0], t[1], t[2], p)) {
				if( !tri_sel_cache.was_preselected( i))
					return i;
			}
		}
	}

	return tri_sel_cache.roll_over_selection();
}

rgVector sk3DRomModel::get_center_of_tv_vertex_cluster( const int *vlist,
	int count)
{
	if( count <= 0) return rgVector();
	rgVector vmin( 16777816.0f, 16777816.0f, 16777816.0f);
	rgVector vmax( -16777816.0f, -16777816.0f, -16777816.0f);
	rgVector u;
	while( count--) {
		u = get_tv_vertex( *(vlist++));
		if( u.x < vmin.x) vmin.x = u.x;
		if( u.y < vmin.y) vmin.y = u.y;
		if( u.z < vmin.z) vmin.z = u.z;
		if( u.x > vmax.x) vmax.x = u.x;
		if( u.y > vmax.y) vmax.y = u.y;
		if( u.z > vmax.z) vmax.z = u.z;
	}
	return (vmin + vmax) * 0.5f;
}

// FIXME: This function does not work yet.
void sk3DRomModel::scale_model( const float x, const float y, const float z)
{
	skel.reset_skel_frame( -1);
	transform_skel_frame();

	//scale_model_skeleton( x, y, z);

	int *vlist = new int[ vertex_count()];
	if( !vlist) return;
	for( int i = 0; i < vertex_count(); i++) {
		rgVector u( get_tv_vertex( i));
		u.x *= x;
		u.y *= y;
		u.z *= z;
		set_tv_vertex( i, u);
		vlist[i] = i;
	}	
	set_vertex_from_tv( vlist, vertex_count());
	delete [] vlist;
}




