/*
Copyright (C) 2004-2005 Amour Hassan
                                                                                
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
                                                                                
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
                                                                                
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
                                                                                
*/

#include "rg3DRom.h"
#include "rgPort.h"
#include <iostream>
#include <fstream>
#include <string.h>
#include <rgOpenGL.h>
#include <rgBSP.h>
#include <rgBSP_template.cpp>

void sk3DRomModel::render_bsp_faces( rgVector p, bool use_normals,
	bool use_textures, const int sid)
{
	render_bsp_faces( p, use_normals, use_textures, sid, 0, 0);
}

void sk3DRomModel::render_bsp_faces( rgVector p, bool use_normals,
	bool use_textures, const int sid, int *flist, const int count)
{
	if( (sid < 0) || (sid >= skel.frame_count())) {
		render_bsp_faces( 0, p, use_normals, use_textures, flist,
			count);
		return;
	}

	skel.m_draw_context.gen_frame_fcs( skel, skel.sf[sid]);
	render_bsp_faces( &(skel.m_draw_context), p, use_normals, use_textures,
		flist, count);
}

void sk3DRomModel::render_bsp_faces( const sk3DRomDrawContext *dc, rgVector p,
	bool use_normals, bool use_textures, int *flist, const int count)
{
	if( bsp_tree.empty()) return;

	// Generate a transformation matrix for the current BSP bone
	// tree based on the bone's orientation (remember that "jcs"
	// is the bone's original orientation, and "jcs" is the
	// orientation that the BSP bone tree was built from).
		
	// NOTE: After the BSP tree was built, all of its face's
	//	vertices where bound to the controlling bone. This
	//	means that a matrix describing "jcs" is required to
	//	to reproduce the original BSP faces and vertices.
	//	In addition, this means that passing "rcs" directly
	//	to the BSP tree renderer will produce the correct
	//	transformed BSP result.
	const int istop = int( bsp_tree.size());
	for( int i = 0; i < istop; i++) {
		rgBSPTree <rgBSP_NTTC_ExtData> *bsp = bsp_tree[i];
		bsp->use_bsp_matrix = true;
		if( dc) {
			bsp->bsp_matrix =
				(dc->get_fcs( bsp_bone_id[i])).
				get_mapping_matrix();
			bsp->bsp_matrix = (dc->mat()) *
				bsp->bsp_matrix;

		} else {
			if( bsp_bone_id[i] >= 0) {
				bsp->bsp_matrix =
					(skel.b[ bsp_bone_id[i]].fcs).
					get_mapping_matrix();

			} else bsp->bsp_matrix.loadIdentity();
		}
		//m = skel.sf[sid].b[ bsp_bone_id[i]].rcs;
			

		bsp->render( p);
		bsp->use_bsp_matrix = false;
	}
}
