/*
Copyright (C) 2004-2005 Amour Hassan
                                                                                
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
                                                                                
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
                                                                                
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
                                                                                
*/

#include "rg3DRom.h"
#include "rgPort.h"
#include <iostream>
#include <string.h>
#include <math.h>



sk3DRomGroup::sk3DRomGroup()
{
	gname[0] = '\0';
	gname[ SK3DROM_GROUP_NAME_MAX_LEN - 1] = '\0';
	gcolor[0] = 0;
	gcolor[1] = 0;
	gcolor[2] = 0;
	gcolor[3] = 0;
	flags = 0;
}

sk3DRomGroup *sk3DRomModel::add_group( const sk3DRomGroup &g)
{
	group.push_back( g);
	return &( group[ group.size() - 1]);
}

inline void hlp_notify_verts_of_group_delete( sk3DRomModel &mdl,
	const int gid)
{
	const int istop = int( mdl.v.size());
	for( int i = 0; i < istop; i++) mdl.remove_vertex_from_group( i, gid);
}

inline void hlp_notify_verts_of_group_swap( sk3DRomModel &mdl,
	const int gid_old, const int gid_new)
{
	const int istop = int( mdl.v.size());
	for( int i = 0; i < istop; i++) {
		if( gid_old == (mdl.v[i].gid()))
			mdl.add_vertex_to_group( i, gid_new);
	}
}

void sk3DRomModel::delete_group( const int gid)
{
	if( gid < 0) return;
	const int last = int( group.size()) - 1;
	if( gid > last) return;

	// The group's "vid" list is cleared as an optimization. This will
	// prevent low-level group handling functions from "bubble-searching"
	// through the group's owned verticies.
	group[ gid].vid.clear();
	hlp_notify_verts_of_group_delete( *this, gid);

	if( gid != last) {
		hlp_notify_verts_of_group_swap( *this, gid, last);
		group[ gid] = group[ last];
	}

	group.erase( group.begin() + last);
}

void sk3DRomModel::build_group_vert_lists()
{
	int i;
	for( i = 0; i < int( group.size()); i++)
		group[i].vid.clear();

	const int istop = int( v.size());
	for( i = 0; i < istop; i++) {
		if( (v[i].gid() >= 0) && (v[i].gid() < int( group.size())))
			group[ v[i].gid()].vid.push_back( i);
	}
}

void sk3DRomModel::remove_vertex_from_group( const int vid, const int gid)
{
	if( v[vid].gid() != gid) return;
	v[ vid].m_gid = -1;
	if( gid < 0) return;
	std::vector <int> &vec = group[gid].vid;
	const int istop = int( vec.size());
	for( int i = 0; i < istop; i++) {
		if( vid == vec[i]) {
			vec.erase( vec.begin() + i);
			return;
		}
	}
}

void sk3DRomModel::add_vertex_to_group( const int vid, const int gid)
{
	remove_vertex_from_group( vid, v[ vid].gid());
	v[ vid].m_gid = gid;
	if( gid < 0) return;
	group[gid].vid.push_back( vid);
}

void sk3DRomModel::set_vertex_gid( int *vlist, int count, const int gid)
{
	if( !vlist) return;
	if( count < 0) return;
	if( gid < 0) return;
	while( count--) add_vertex_to_group( *(vlist++), gid);
}

void sk3DRomModel::hide_group( const int gid)
{
	if( (gid < 0) || (gid > int( group.size()))) return;
	group[gid].flags |= SK3DROM_GROUP_HIDDEN_BIT;
	std::vector <int> &vid = group[gid].vid;
	const int istop = int( vid.size());
	for( int i = 0; i < istop; i++) v[ vid[i]].hide();
}

void sk3DRomModel::reveal_group( const int gid)
{
	if( (gid < 0) || (gid > int( group.size()))) return;
	group[gid].flags &= ~SK3DROM_GROUP_HIDDEN_BIT;
	std::vector <int> &vid = group[gid].vid;
	const int istop = int( vid.size());
	for( int i = 0; i < istop; i++) v[ vid[i]].reveal();
}

const int *sk3DRomModel::get_faces_in_group( const int gid, int &count)
{
	count = 0;
	if( m_face_list) { delete m_face_list; m_face_list = 0;}
	if( gid < 0) return 0;

	{ // Determine total number of faces in group.
		for( int i = 0; i < face_count(); i++)
			if( gid == f[i].gid) count++;
	}

	if( count <= 0) return 0;
	if( !(m_face_list = new int[ count])) return 0;
	
	int *p = m_face_list;
	for( int i = 0; i < face_count(); i++)
		if( gid == f[i].gid) *(p++) = i;

	return m_face_list;
}

int sk3DRomModel::get_group_id( const char *gname) const
{
	if( !gname) return -1;
	for( int i = 0; i < group_count(); i++)
		if( 0 == strncmp( gname, group[i].gname,
			SK3DROM_GROUP_NAME_MAX_LEN)) return i;
	return -1;
}

