/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "camera.h"
#include "engine/engine_game.h"
#include "engine/engine_graphics.h"
//#include <rg3DRom.h>

std::map <const std::string, int (*)( const gmCCArg &)> phCamera::m_ccmd;
std::map <const std::string, int (*)( const gmCCArg &)> phCameraPitch::m_ccmd;

extern "C" gmThing *spw_phCamera()
{ return new phCamera();}

extern "C" int del_phCamera( gmThing *t)
{ delete t; return 0;}

extern "C" gmThing *spw_phCameraPitch()
{ return new phCameraPitch();}

extern "C" int del_phCameraPitch( gmThing *t)
{ delete t; return 0;}

extern "C" int gm_main( const std::string *key, GameEngine *ge,
	std::string *emsg)
{
	ge->m_thing_manager->add_to_spawn_list( "phCamera", spw_phCamera);
	ge->m_thing_manager->add_to_delete_list( "phCamera", del_phCamera);

	ge->m_thing_manager->add_to_spawn_list(
		"phCameraPitch", spw_phCameraPitch);
	ge->m_thing_manager->add_to_delete_list(
		"phCameraPitch", del_phCameraPitch);

	return 0;
}

phCamera::phCamera()
{
	if( !(m_ccmd.empty())) return;
	add_console_command( "target", phCamera_target, m_ccmd);
}

phCamera::~phCamera()
{
}

int phCamera_target( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) {
		p.ge->rcout( "usage: target <thing>\n"
			"NOTE: \"thing\" is the unique name of the thing to "
				"target.\n");
		return GM_THING_ERR_PARAM;
	}
	ThingControlBlock *t = p.ge->m_thing_manager->get( args[0]);
	if( !t) {
		p.ge->rcout( "No such thing \"" + args[0] + "\"\n");
		return GM_THING_ERR_GENERIC;
	}
	((phCamera *)(p.thing))->m_look_at_target = t;
	return 0;
}

int phCamera::call_ccmd( const std::string &key, gmCCArg &args)
{
	std::map <const std::string, int (*)( const gmCCArg &)> ::
		const_iterator iter = m_ccmd.find( key);
	if( iter == m_ccmd.end()) return gmModelThing::call_ccmd( key, args);
	return (iter->second)( args);
}

std::string phCamera::list_ccmd( const std::string &p_prefix)
{
	std::map <const std::string, int (*)( const gmCCArg &)> ::
		const_iterator iter = m_ccmd.begin();
	std::string tmp;
	for( ; iter != m_ccmd.end(); iter++) {
		tmp += p_prefix + phCamera::type() + "::" + (iter->first) +
			"\n";
	}
	tmp += gmModelThing::list_ccmd( p_prefix);
	return tmp;
}

void phCamera::process( ThingManager *tm)
{
}




// XXX: Pitch camera
phCameraPitch::phCameraPitch()
{
	if( !(m_ccmd.empty())) return;
	add_console_command( "heading", phCameraPitch_heading, m_ccmd);
}

int phCameraPitch::call_ccmd( const std::string &key, gmCCArg &args)
{
	std::map <const std::string, int (*)( const gmCCArg &)> ::
		const_iterator iter = m_ccmd.find( key);
	if( iter == m_ccmd.end()) return phCamera::call_ccmd( key, args);
	return (iter->second)( args);
}

std::string phCameraPitch::list_ccmd( const std::string &p_prefix)
{
	std::map <const std::string, int (*)( const gmCCArg &)> ::
		const_iterator iter = m_ccmd.begin();
	std::string tmp;
	for( ; iter != m_ccmd.end(); iter++) {
		tmp += p_prefix + phCameraPitch::type() + "::" + (iter->first)
			+ "\n";
	}
	tmp += phCamera::list_ccmd( p_prefix);
	return tmp;
}

int phCameraPitch_heading( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 3) {
		p.ge->rcout( "usage: heading <x> <y> <z>\n");
		return GM_THING_ERR_PARAM;
	}
	((phCameraPitch *)(p.thing))->m_heading = gm_str_to_vector( args[0],
		args[1], args[2]);
	return 0;
}

void phCameraPitch::process( ThingManager *tm)
{
	gmCamera *m_cam = controlled_camera();
	if( !m_cam) return;

	gmThing *tar = get_look_at_thing();
	if( !tar) return;

//	m_cam->move_to( (tar->cs()).p - (m_heading * 10.0f), 0.25f);
	m_cam->look_at( (tar->cs()).p, rgVector( 0, 1, 0));
//	const rgVector p( m_cam->cs.p);
//	m_cam->cs = cs();
//	m_cam->cs.p = p;
}

