/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <SDL/SDL.h>
#include "main_header.h"
#include "engine_game.h"

int krn_initialize_system()
{
	if( SDL_Init( SDL_INIT_VIDEO)) { SDL_Quit(); return -1;}
	return 0;
}

int krn_choose_good_video_mode( MainData &mdata)
{
	for( int i = 0; i < int( mdata.m_video_mode.size()); i++) {
		const gmVideoInfo &vi = mdata.m_video_mode[i];
		SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, vi.zbuffer);
		if( SDL_VideoModeOK(
			vi.width,
			vi.height,
			vi.bpp,
			SDL_OPENGL | SDL_HWSURFACE)) return i;
	}
	return -1;
}

int krn_create_screen( GameEngine &ge)
{
	SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 8);
	SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 8);
	SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 8);
	SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1);

	const int vid = krn_choose_good_video_mode( *(ge.mdata));
	if( vid < 0) return -1;

	const gmVideoInfo &vi = ge.mdata->m_video_mode[ vid];
	ge.rcout( "Screen Settings (choice " + gm_to_string( vid + 1) + " of "
		+ gm_to_string( int( ge.mdata->m_video_mode.size())) +
		"):\n\twidth = " + gm_to_string( vi.width) +
		"\n\theight = " + gm_to_string( vi.height) +
		"\n\tcolour depth (bpp) = " + gm_to_string( vi.bpp) +
		"\n\tzbuffer depth = " + gm_to_string( vi.zbuffer) + "\n");

	ge.mdata->m_screen = SDL_SetVideoMode(
		vi.width,
		vi.height,
		vi.bpp,
		SDL_OPENGL | SDL_HWSURFACE);
	if( NULL == (ge.mdata->m_screen)) return -2;
	return 0;
}
