/*
Copyright (C) 2004-2005 Amour Hassan
                                                                                
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
                                                                                
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
                                                                                
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
                                                                                
*/

#include "things.h"
#include "engine/engine_game.h"
#include "engine/engine_graphics.h"
#include <fstream>
#include <sstream>
#include <rg3DRom.h>

std::map <const std::string, int (*)( const gmCCArg &)> gmModelThing::m_ccmd;

extern "C" gmThing *spw_model_thing()
{
	return new gmModelThing();
}

extern "C" int del_model_thing( gmThing *t)
{
	delete t;
	return 0;
}

gmModelThing::gmModelThing()
{
	m_mh = 0;
	m_draw_context = new sk3DRomDrawContext();

	m_enable_animation = false;
	m_ani_clock_prev = 0;

	// XXX: When using vertex arrays, it is best to cause the model to
	//	render vertices directly to the draw context. If this is not
	//	desired, then the "thing" can manually disable it.
	m_draw_context->m_transform_direct = true;

	if( !(m_ccmd.empty())) return;
	add_console_command( "set_model", gmModelThing_set_model, m_ccmd);
	add_console_command( "transform", gmModelThing_transform, m_ccmd);
	add_console_command( "dump_matrix", gmModelThing_dump_matrix, m_ccmd);
	add_console_command( "load_identity", gmModelThing_load_identity,
		m_ccmd);
	add_console_command( "scale", gmModelThing_scale, m_ccmd);
	add_console_command( "translate", gmModelThing_translate, m_ccmd);
	add_console_command( "rotate", gmModelThing_rotate, m_ccmd);
	add_console_command( "animate", gmModelThing_animate, m_ccmd);
	add_console_command( "set_dc_root_master",
		gmModelThing_set_dc_root_master, m_ccmd);
	add_console_command( "clear_display_lists",
		gmModelThing_clear_display_lists, m_ccmd);
	add_console_command( "create_display_list",
		gmModelThing_create_display_list, m_ccmd);
	add_console_command( "gen_display_list",
		gmModelThing_gen_display_list, m_ccmd);
	add_console_command( "use_display_list", gmModelThing_use_display_list,
		m_ccmd);
	add_console_command( "use_vertex_array", gmModelThing_use_vertex_array,
		m_ccmd);
}

gmModelThing::~gmModelThing()
{
	if( clear_display_lists()) {
		std::cerr << "~gmModelThing(): Thing \"" << get_key() <<
			"\": Could not clear display lists.\n"
			"\tLists remaining: " << m_display_list.size() << "\n";
	}
	if( m_draw_context) delete m_draw_context;
	unreference_model();
}

int gmModelThing::call_ccmd( const std::string &key, gmCCArg &args)
{
	std::map <const std::string, int (*)( const gmCCArg &)> ::
		const_iterator iter = m_ccmd.find( key);
	if( iter == m_ccmd.end())
		return gmThing::call_ccmd( key, args);
	return (iter->second)( args);
}

std::string gmModelThing::list_ccmd( const std::string &p_prefix)
{
	std::map <const std::string, int (*)( const gmCCArg &)> ::
		const_iterator iter = m_ccmd.begin();
	std::string tmp;
	for( ; iter != m_ccmd.end(); iter++) {
		tmp += p_prefix + gmModelThing::type() + "::" + (iter->first) +
			"\n";
	}
	tmp += gmThing::list_ccmd( p_prefix);
	return tmp;
}


int gmModelThing_transform( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	switch( args.size()) {
	case 0: ((gmModelThing *)(p.thing))->transform(); return 0;
	case 2:
		if( "int" == args[0]) {
			((gmModelThing *)(p.thing))->transform( atoi(
				(args[1]).c_str()));

		} else if( "key" == args[0]) {
			((gmModelThing *)(p.thing))->transform( args[1]);

		} else {
			p.ge->rcout( "gmModelThing_transform:\n"
				"\tusage: transform [\"int\"/\"key\"] "
					"[integer/string]\n"
				"\texample 1: transform\n"
				"\texample 2: transform int 5\n"
				"\texample 3: transform key \"idle_frame\"\n");
			return -1;
		}
		return 0;

	default: return -2;
	}
}

int gmModelThing_dump_matrix( const gmCCArg &p)
{
	gmModelThing *mt = (gmModelThing *)(p.thing);
	if( !mt->m_draw_context) return GM_THING_ERR_GENERIC;
	const std::vector <std::string> &args = *(p.args);
	if( args.size() <= 0) {
		p.ge->rcout( "usage: dump_matrix [which matrix] "
			"[which matrix] [which matrix] ...\n"
			"NOTE: Available matrices are:\n"
			"\t\"scale\", \"translate\", \"rotate\", "
			"\"transformation\", \"all\"\n"
			"Example 1: dump_matrix \"all\"\n"
			"Example 2: dmp_matrix \"translate\"\n");
		return GM_THING_ERR_GENERIC;
	}
	for( int i = 0; i < int( args.size()); i++) {
		if( "all" == args[i]) {
			std::ostringstream des;
			des	<< "Transformation\n"
					<< mt->m_draw_context->mat()
				<< "\nScale\n"
					<< mt->m_draw_context->m_scale
				<< "\nTranslation\n"
					<< mt->m_draw_context->m_translate
				<< "\nRotation\n"
					<< mt->m_draw_context->m_rotate
				<< "\n";
			p.ge->rcout( des.str());

		} else if( "scale" == args[i]) {
			std::ostringstream des;
			des << "Scale\n" << mt->m_draw_context->m_scale
				<< "\n";
			p.ge->rcout( des.str());

		} else if( "translate" == args[i]) {
			std::ostringstream des;
			des << "Translation\n"
				<< mt->m_draw_context->m_translate << "\n";
			p.ge->rcout( des.str());

		} else if( "rotate" == args[i]) {
			std::ostringstream des;
			des << "Rotation\n"
				<< mt->m_draw_context->m_rotate << "\n";
			p.ge->rcout( des.str());

		} else if( "transformation" == args[i]) {
			std::ostringstream des;
			des << "Transformation\n"
				<< mt->m_draw_context->mat() << "\n";
			p.ge->rcout( des.str());

		} else p.ge->rcout( "No such matrix \"" + args[i] + "\"\n");
	}
	return 0;
}

int gmModelThing_load_identity( const gmCCArg &p)
{
	gmModelThing *mt = (gmModelThing *)(p.thing);
	if( !mt->m_draw_context) return GM_THING_ERR_GENERIC;
	const std::vector <std::string> &args = *(p.args);
	if( args.size() <= 0) {
		p.ge->rcout( "usage: load_identity [which matrix] "
			"[which matrix] [which matrix] ...\n"
			"NOTE: Available matrices are:\n"
			"\t\"scale\", \"translate\", \"rotate\", "
			"\"transformation\", \"all\"\n"
			"Example 1: load_identity \"all\"\n"
			"Example 2: load_identity \"translate\"\n");
		return GM_THING_ERR_GENERIC;
	}
	for( int i = 0; i < int( args.size()); i++) {
		if( "all" == args[i]) {
//			mt->m_draw_context->m_mat.loadIdentity();
			mt->m_draw_context->m_scale.loadIdentity();
			mt->m_draw_context->m_translate.loadIdentity();
			mt->m_draw_context->m_rotate.loadIdentity();

		} else if( "scale" == args[i]) {
			mt->m_draw_context->m_scale.loadIdentity();

		} else if( "translate" == args[i]) {
			mt->m_draw_context->m_translate.loadIdentity();

		} else if( "rotate" == args[i]) {
			mt->m_draw_context->m_rotate.loadIdentity();

		} else if( "transformation" == args[i]) {
//			mt->m_draw_context->m_mat.loadIdentity();

		} else p.ge->rcout( "No such matrix \"" + args[i] + "\"\n");
	}
//	mt->m_draw_context->update_matrix();
	return 0;
}

int gmModelThing_set_dc_root_master( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) {
		p.ge->rcout( "usage: set_dc_root_master <key>\n"
			"NOTE: \"key\" is the name of the bone to make the "
			"root master.\n");
		return GM_THING_ERR_PARAM;
	}
	gmModelThing *mt = (gmModelThing *)(p.thing);
	if( !mt->m_draw_context) return GM_THING_ERR_GENERIC;
	if( !mt->m_mh) {
		p.ge->rcout( "Can't set root master, no model is set\n");
		return GM_THING_ERR_GENERIC;
	}
	const int rmid = mt->m_mh->bone_id( args[0]);
	if( rmid < 0) {
		p.ge->rcout( "Attached model has no bone named \"" + args[0] +
			"\"\n");
		return GM_THING_ERR_GENERIC;
	}
	mt->m_draw_context->set_root_master( rmid);
	return 0;
}

int gmModelThing_scale( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( (args.size() != 3) && (args.size() != 4)) {
		p.ge->rcout( "gmModelThing_scale: <x> <y> <z> [\"pre\"]\n"
			"NOTE: \"pre\" means to scale the pre-scaling "
				"matrix.\n");
		return GM_THING_ERR_PARAM;
	}
	gmModelThing *mt = (gmModelThing *)(p.thing);
	if( !mt->m_draw_context) return GM_THING_ERR_GENERIC;
	const rgVector v( gm_str_to_vector( args[0], args[1], args[2]));
	if( 4 == args.size()) {
		if( "pre" == args[3]) mt->m_draw_context->pre_scale( v);
		else {
			p.ge->rcout( "Bad option \"" + args[3] + "\"\n");
			return GM_THING_ERR_PARAM;
		}
	} else mt->m_draw_context->scale( v);
//	mt->m_draw_context->update_matrix();
	return 0;
}

int gmModelThing_translate( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( (args.size() != 3) && (args.size() != 4)) {
		p.ge->rcout( "gmModelThing_translate: <x> <y> <z> [\"pre\"]\n"
			"NOTE: \"pre\" means to translate the pre-translation "
				"matrix.\n");
		return GM_THING_ERR_PARAM;
	}
	gmModelThing *mt = (gmModelThing *)(p.thing);
	if( !mt->m_draw_context) return GM_THING_ERR_GENERIC;
	const rgVector v(  gm_str_to_vector( args[0], args[1], args[2]));
	if( 4 == args.size()) {
		if( "pre" == args[3]) mt->m_draw_context->pre_translate( v);
		else {
			p.ge->rcout( "Bad option \"" + args[3] + "\"\n");
			return GM_THING_ERR_PARAM;
		}
	} else mt->m_draw_context->translate( v);
//	mt->m_draw_context->update_matrix();
	return 0;
}

int gmModelThing_rotate( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( (args.size() != 4) && (args.size() != 5)) {
		p.ge->rcout( "gmModelThing_rotate: <ang> <axis vector> "
				"[\"pre\"]\n"
			"NOTE: \"ang\" is given in degrees.\n"
			"The \"pre\" option means rotate the pre-rotation "
				"matrix.\n");
		return GM_THING_ERR_PARAM;
	}
	gmModelThing *mt = (gmModelThing *)(p.thing);
	if( !mt->m_draw_context) return GM_THING_ERR_GENERIC;
	const float ang =  RG_DEGTORAD( strtod( (args[0]).c_str(), 0));
	const rgVector axis( (gm_str_to_vector( args[1], args[2], args[3])).
		normalize());

	if( 5 == args.size()) {
		if( "pre" == args[4])
			mt->m_draw_context->pre_rotate( ang, axis);
		else {
			p.ge->rcout( "Bad option \"" + args[3] + "\"\n");
			return GM_THING_ERR_PARAM;
		}
	} else mt->m_draw_context->rotate( ang, axis);
//	mt->m_draw_context->update_matrix();
	return 0;
}

int gmModelThing_set_model( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) return GM_THING_ERR_PARAM;

	gmModelThing *m = (gmModelThing *)(p.thing);
	if( !(m->reference_model( args[0], *(p.ge->gfx_engine->model_man()))))
		return GM_THING_ERR_GENERIC;
	(m->model())->skel.init_skel_frame( m->m_draw_context->m_skout);
	m->m_draw_context->init_fcs( (m->model())->skel);
	p.ge->rcout( "model for thing \"" + ((p.thing)->tcb())->m_key +
		"\" is now \"" + args[0] + "\"\n");
	return 0;
}

int gmModelThing_use_vertex_array( const gmCCArg &p)
{
	gmModelThing *m = (gmModelThing *)(p.thing);
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) {
		m->m_draw_context->m_use_vertex_array = true;
		return 0;
	}
	gm_set_state( m->m_draw_context->m_use_vertex_array, args[0]);
	return 0;
}

int gmModelThing_clear_display_lists( const gmCCArg &p)
{
	((gmModelThing *)(p.thing))->clear_display_lists();
	return 0;
}

int gmModelThing_create_display_list( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) {
		p.ge->rcout( "usage: create_display_list <key>\n");
		return GM_THING_ERR_PARAM;
	}

	gmModelThing *m = (gmModelThing *)(p.thing);
	if( !(m->create_display_list( args[0]))) {
		p.ge->rcout( "Could not create display list \"" + args[0] +
			"\"\n");
		return GM_THING_ERR_GENERIC;
	}
	return 0;
}

int gmModelThing_gen_display_list( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) {
		p.ge->rcout( "usage: gen_display_list <key>\n");
		return GM_THING_ERR_PARAM;
	}

	gmModelThing *m = (gmModelThing *)(p.thing);
	const int err = m->gen_display_list( args[0]);
	if( err) {
		p.ge->rcout( "Could not generate display list \"" + args[0] +
			"\"\n");
		return GM_THING_ERR_GENERIC;
	}
	return 0;
}

int gmModelThing_use_display_list( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() > 1) {
		p.ge->rcout( "usage: use_display_list [key]\n");
		return GM_THING_ERR_PARAM;
	}

	gmModelThing *m = (gmModelThing *)(p.thing);
	if( 0 == args.size()) {
		m->m_use_display_list = std::string();
		return 0;
	}
	m->m_use_display_list = args[0];
	return 0;
}


#define TMP_NO_MODEL	-1
#define TMP_NO_SEQUENCE	-2
int gmModelThing_animate( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	gmModelThing *m = (gmModelThing *)(p.thing);
	if( 1 == args.size()) {
		bool b = m->m_enable_animation;
		m->enable_animation( gm_set_state( b, args[0]));
		return 0;

	} else if( args.size() != 4) {
		p.ge->rcout( "usage 1: animate <0/1/on/off/toggle>\n"
			"usage 2: animate <sequence key> <frame index> "
				"<initial transition time> <elapsed time>\n"
			"NOTE: \"sequence key\" is the unique name that "
				"identifies the sequence.\n"
			"To start an animation immediatly, pass 1 and 1 for "
				"both transition and elapsed time.\n");
		return -1;
	}

	const int err = m->animate( args[0], atoi( (args[1]).c_str()),
		atoi( (args[2]).c_str()), atoi( (args[3]).c_str()));
	switch( err) {
	case 0: break;
	case TMP_NO_MODEL:
		p.ge->rcout( "Thing \"" + m->get_key() +
			"\" has no model to animate\n");
		return -2;

	case TMP_NO_SEQUENCE:
		p.ge->rcout( "Thing \"" + m->get_key() +
			"\" has no sequence named \"" + args[0] + "\"\n");
		return -3;

	default:
		p.ge->rcout( "Thing \"" + m->get_key() + "\": Error " +
			gm_to_string( err) + "while animating \"" + args[0] +
			"\"\n");
		return -4;
	}
	return 0;
}

int gmModelThing::animate( const std::string &key, int fid, const int p_time,
	const int p_elapsed)
{
	if( !model()) return TMP_NO_MODEL;
	const int sid = m_mh->sequence_id( key);
	if( sid < 0) return TMP_NO_SEQUENCE;
	(model())->skel.animate( m_mh->sequence_id( key), fid, p_time,
		p_elapsed, m_draw_context);
	return 0;
}
#undef TMP_NO_MODEL
#undef TMP_NO_SEQUENCE

void gmModelThing::animate( const int p_elapsed)
{
	if( !model()) return;
	(model())->skel.animate( p_elapsed, m_draw_context);
}

void gmModelThing::animate()
{
	if( !m_enable_animation) return;
	long long mili_elapsed = ((ge())->game_clock() - m_ani_clock_prev) *
		(ge())->game_clock_tick();
	if( 0 == mili_elapsed) return;
	m_ani_clock_prev = (ge())->game_clock();
	animate( mili_elapsed);
}

bool gmModelThing::reference_model( const std::string &key, ModelManager &m)
{
	ModelHandle *h = m.get( key);
	if( !h) return false;
	unreference_model();
	m_mh = h;
	m_mh->inc_reference();
	return true;
}
                                                                                
void gmModelThing::unreference_model()
{
	if( !m_mh) return;
	m_mh->dec_reference();
	m_mh = 0;
}

sk3DRomModel *gmModelThing::model()
{
	if( !m_mh) return 0;
	return m_mh->model;
}

void gmModelThing::draw( gmRenderArgs *ra)
{
	if( m_use_display_list.length() > 0) {
		draw_display_list( m_use_display_list);
		return; 
	}

	if( !(model())) return;
	if( !ra) return;

//	bool shadow = false;
//	if( ra) shadow = ra->shadow;
	
	// FIXME: It may not be neccasary to always transform the model
	// (for example, it may have been previously transformed, so another
	// transformation is redundant). Add a check to prevent redundant
	// transformations.
	transform();

	glColor3f( 1, 1, 1);
	if( ra->render_opaque) {
		if( m_draw_context->m_use_vertex_array) {
			if( !((model())->get_vertex_array()))
				(model())->create_vertex_array();
			else (model())->update_vertex_array();
//			else if( !m_draw_context->m_transform_direct)
//				(model())->update_vertex_array();
			(model())->render_vertex_array( m_draw_context);

		} else {
			(model())->render_faces( true, true);
		}
	}

	if( ra->render_transparent) {
		gmCamera *c = (ge())->gfx_engine->current_camera();
		if( c) {
			(model())->render_bsp_faces( m_draw_context, c->cs.p,
				true, true, 0, 0);
		}
	}
}

void gmModelThing::enable_animation( const bool b)
{
	m_enable_animation = b;
	if( b) m_ani_clock_prev = (ge())->game_clock();
}

void gmModelThing::transform()
{
	if( !(model())) return;
	(model())->transform_skel_frame( m_draw_context);
}

void gmModelThing::transform( int fid)
{
	if( !(model())) return;
	(model())->transform_skel_frame( fid, m_draw_context);
}

int gmModelThing::frame_id( const std::string &key) const
{
	if( !m_mh) return -1;
	return m_mh->frame_id( key);
}


/*
void gmModelThing::transform( int src, int des, float weight)
{
	if( !(model())) return;
	if( !m_skel) {
		if( !(m_skel = new sk3DRomSkelFrame())) return;
		(model())->skel->init_skel_frame( *m_skel);
	}

	(model())->skel->inbetween_skeleton( weight, src, des, *m_skel);
}
*/

GLuint gmModelThing::create_display_list( const std::string &key)
{
	typedef std::pair <const std::string, GLuint> SIPair;
	GLuint tmp = glGenLists( 1);
	if( 0 == tmp) return 0;
	if( false == ((m_display_list.insert( SIPair( key, tmp))).second)) {
		glDeleteLists( tmp, 1);
		return 0;
	}
	return tmp;
}

GLuint gmModelThing::get_display_list( const std::string &key) const
{
	std::map <const std::string, GLuint> :: const_iterator iter =
		m_display_list.find( key);
	if( iter == m_display_list.end()) return 0;
	return (iter->second);
}

int gmModelThing::delete_display_list( const std::string &key)
{
	std::map <const std::string, GLuint> :: iterator iter =
		m_display_list.find( key);
	if( iter == m_display_list.end()) return -1;
	glDeleteLists( iter->second, 1);
	m_display_list.erase( iter);
	return 0;
}

int gmModelThing::clear_display_lists()
{
	while( !(m_display_list.empty())) {
		std::map <const std::string, GLuint> :: iterator iter =
			m_display_list.begin();
		if( delete_display_list( iter->first)) return -1;
	}
	return 0;
}

int gmModelThing::gen_display_list( const std::string &key)
{
	const GLuint tmp = get_display_list( key);
	if( !tmp) return -1;
	m_draw_context->backup_matrices();
	m_draw_context->reset_all_matrices();
	glNewList( tmp, GL_COMPILE);
	gmRenderArgs tmp_ra;
	gmModelThing::draw( &tmp_ra);
	glEndList();
	m_draw_context->restore_matrices();
	return 0;
}

int gmModelThing::draw_display_list( const std::string &key) const
{
	const GLint tmp = get_display_list( key);
	if( !tmp) return -1;
	glPushMatrix();
	rgMatrix m;
	getRGMatrix( m);
	m *= m_draw_context->mat();
	loadRGMatrix( m);
	glCallList( tmp);
	glPopMatrix();
	return 0;
}

void gmModelThing::render_face_group( const std::string &grp_name,
	const bool use_normals, const bool use_textures)
{
	if( !(model())) return;
	int count;
	const int *flist = (model())->get_faces_in_group(
		(model())->get_group_id( (grp_name).c_str()), count);
	if( flist) {
		(model())->render_faces( use_normals, use_textures, flist,
			count);
	}
}

void gmModelThing::process( ThingManager *tm)
{
	animate();
}
